#!/bin/sh
set -efu

### Tools ###

CONFIGURE_ARGS=${CONFIGURE_ARGS-"-C --enable-maintainer-mode"}
MAKE=${MAKE-"make"}
MAKE_FLAGS=${MAKE_FLAGS-""}
CPP=${CPP-"cpp"}
CPP_FLAGS=${CPP_FLAGS-""}

### global variables ###

cvs_repository=""
cvs_module="mc"
dir_basedir=""
dir_srcdir=""
#dir_workdir (will be set later)
#dir_prefix (will be set later)
#dir_logdir (will be set later)

configure_args_enable_all="--enable-largefile --enable-nls "\
"--enable-netcode --enable-background --enable-charset "\
"--with-mmap --with-included-gettext --with-x "\
"--with-vfs --with-mcfs --with-samba --with-ext2undel "\
"--with-gpm-mouse --with-subshell --with-edit"
configure_args_disable_all="--disable-largefile --disable-nls "\
"--disable-netcode --disable-background --disable-charset "\
"--without-mmap --without-included-gettext --without-x "\
"--without-vfs --without-mcfs --without-samba --without-ext2undel "\
"--without-gpm-mouse --without-subshell --without-edit"
gcc_flags_warns1="-O2 -Wall -Wno-unused-parameter -Wno-sign-compare "\
"-Wmissing-declarations -Wmissing-prototypes -Wbad-function-cast "\
"-Wcast-align -Wpointer-arith -Wredundant-decls -Wundef -Wfloat-equal"
gcc_flags_warns2="${gcc_flags_warns1} -W -Wwrite-strings"

do_cleanup=no
do_cleansrc=no
do_preconfigure=no
do_extract=no
do_cleanwork=yes
do_configure=yes
do_make_clean=yes
do_make_all=yes
do_make_check=yes
do_make_install=yes

nice_value="19"
enable_gcc_warnings=yes
#quiet (default: undefined)

errors=0
warnings=0


usage() {
	cat >&2 <<EOF
usage: `basename $0` [options ...] [<testname ...>]

	-basedir <dir>		set the base directory
	-cvs-repository <repos>	select the CVS repository to extract from
	-cvs-module <module>	select the CVS module (default: ${cvs_module})
	-srcdir <dir>		select the directory where the source code is
				read from or will be extracted to from CVS

	-[dont-]cleansrc	recreate the source directory (default: ${do_cleansrc})
	-[dont-]preconfigure	generate the configure script (default: ${do_preconfigure})
	-[dont-]cleanwork	recreate the build directory (default: ${do_cleanwork})
	-[dont-]configure	run configure (default: ${do_configure})
	-[dont-]clean		make clean (default: ${do_make_clean})
	-[dont-]build		make all (default: ${do_make_all})
	-[dont-]check		make check (default: ${do_make_check})
	-[dont-]install		make install (default: ${do_make_install})

	-fast			only do what's necessary
	-nice <n>		set the nice(1) value (default: ${nice_value})
	-quiet			disable most status messages (default: ${quiet:-"no"}${quiet+"yes"})
	-enable-gcc-warnings	enable warning flags for GCC (default)
	-disable-gcc-warnings	disable warning flags for GCC
EOF
	exit 1
}

show_file() {
	if test -f "$1"; then
		echo "" 1>&2
		sed -e '10s,.*,(more lines follow ...),' -e '10q' "$1" 1>&2
		echo "" 1>&2
	fi
	return 0
}

phase_start() {
	echo -n "[`date`] $1"
	return 0
}

phase_ok() {
	echo " ok"
	return 0
}

phase_warnings() {
	echo " WARNINGS"
	warnings=`expr $warnings + 1`
	if test $# -gt 0; then show_file "$1"; fi
	return 0
}

phase_ok_or_warnings() {
	fsize=`wc -c < "$1"`

	{ test -n "${fsize}" && test ${fsize} -eq 0
	} && phase_ok || phase_warnings "$1"
	return 0
}

phase_failed() {
	echo " FAILED"
	failed=yes
	errors=`expr $errors + 1`
	if test $# -gt 0; then show_file "$1"; fi
	return 0
}

setup_testenv=\
"test_srcdir=\"\${dir_srcdir}\";"\
"test_workdir=\"\${dir_workdir}/\${testname}\";"\
"test_prefix=\"\${dir_prefix}/\${testname}\";"\
"test_logdir=\"\${dir_logdir}/\${testname}\";"\
"out=\"\${test_logdir}/\${test_phase}.out\";"\
"err=\"\${test_logdir}/\${test_phase}.err\";"

confbuild_cleansrc() {
	test_phase="cleansrc"
	eval "${setup_testenv}"

	test ${do_cleansrc} = yes || return 0

	phase_start "cleaning source directory for ${testname} ..."
	{ nice -${nice_value} rm -rf "${test_srcdir}" "${test_workdir}" "${test_prefix}" "${test_logdir}"
	} && phase_ok || phase_failed
}

confbuild_extract() {
	test_phase="extract"
	eval "${setup_testenv}"

	if test ${do_extract} = no && test -d "${test_srcdir}/CVS"; then
		return 0
	fi

	mkdir -p "${test_logdir}"
	if test ${quiet-"no"} = yes; then
		cvs_quiet="-Q"
	else
		cvs_quiet="-q"
	fi

	if test -d "${test_srcdir}/CVS"; then
		phase_start "updating CVS copy for ${testname} ..."
		( cd "${test_srcdir}" \
		  && nice -${nice_value} cvs -z3 ${cvs_quiet} update -dP
		) 1>"${out}" 2>"${err}" && phase_ok_or_warnings "${err}" || phase_failed "${err}"
	else
		if test -z "${cvs_repository}"; then
			echo "CVS repository unset -- the -cvs-repository option is required." 1>&2
			exit 1
		fi
		phase_start "getting fresh CVS copy for ${testname} ..."
		( mkdir -p "${test_srcdir}" \
		  && cd "${test_srcdir}/.." \
		  && nice -${nice_value} cvs -z3 ${cvs_quiet} -d "${cvs_repository}" checkout -P -d "`basename "${test_srcdir}"`" "${cvs_module}"
		) 1>"${out}" 2>"${err}" && phase_ok_or_warnings "${err}" || phase_failed "${err}"
	fi
}

confbuild_preconfigure() {
	test_phase="preconfigure"
	eval "${setup_testenv}"

	if test ${do_preconfigure} = no && test -f "${test_srcdir}/configure"; then
		return 0
	fi

	if test ${do_preconfigure} = once; then
		do_preconfigure=no
	fi

	mkdir -p "${test_logdir}"

	phase_start "preparing configuration for ${testname} ..."
	{ cd "${test_srcdir}" \
	  && echo "#!/bin/sh" > ./configure.mc \
	  && chmod +x ./configure.mc \
	  && nice -${nice_value} ${SHELL-"/bin/sh"} ./autogen.sh
	} 1>"${out}" 2>"${err}" && phase_ok_or_warnings "${err}" || phase_failed "${err}"
}

confbuild_cleanwork() {
	test_phase="cleanwork"
	eval "${setup_testenv}"

	if test ${do_cleanwork} = no || test ! -d "${test_workdir}"; then
		return 0
	fi

	mkdir -p "${test_logdir}"

	phase_start "cleaning directory for ${testname} ..."
	{ rm -rf "${test_workdir}"
	} 1>"${out}" 2>"${err}" && phase_ok_or_warnings "${err}" || phase_failed "${err}"
}

confbuild_configure() {
	test_phase="configure"
	eval "${setup_testenv}"

	if test ${do_configure} = no && test -f "${test_workdir}/Makefile"; then
		return 0
	fi

	mkdir -p "${test_logdir}"

	phase_start "configuring for ${testname} ..."
	( mkdir -p "${test_workdir}" \
	  && cd "${test_workdir}" \
	  && { test ${enable_gcc_warnings} = yes && test ! -z "${gcc_flags-""}" && cflags="CFLAGS=\"${gcc_flags}\"" || cflags=""; } \
	  && eval "nice -${nice_value} ${test_srcdir}/configure --prefix=${test_prefix} ${cflags} ${quiet+"-q"} ${configure_args-""} ${CONFIGURE_ARGS}"
	) 1>"${out}" 2>"${err}" && phase_ok_or_warnings "${err}" || phase_failed "${err}"
}

confbuild_make() {
	make_target="$1"
	test_phase="make_${make_target}"
	eval "${setup_testenv}"

	if eval "test \${do_make_${make_target}} = no"; then
		return 0
	fi

	mkdir -p "${test_logdir}"

	phase_start "running \"make ${make_target}\" for ${testname} ..."
	( cd "${test_workdir}" \
	  && eval "nice -${nice_value} ${MAKE} ${quiet+"-s"} ${MAKE_FLAGS} ${make_target}"
	) 1>"${out}" 2>"${err}" && phase_ok_or_warnings "${err}" || phase_failed "${err}"
}

confbuild() {
	failed=no

	confbuild_cleansrc
	if test ${failed} = yes; then return 0; fi

	confbuild_extract
	if test ${failed} = yes; then return 0; fi

	confbuild_preconfigure
	if test ${failed} = yes; then return 0; fi

	confbuild_cleanwork
	if test ${failed} = yes; then return 0; fi

	confbuild_configure
	if test ${failed} = yes; then return 0; fi

	confbuild_make "clean"
	if test ${failed} = yes; then return 0; fi

	confbuild_make "all"
	if test ${failed} = yes; then return 0; fi

	confbuild_make "check"
	if test ${failed} = yes; then return 0; fi

	confbuild_make "install"
	if test ${failed} = yes; then return 0; fi
}


test_plain() {
	testname="plain" \
	configure_args="" \
	confbuild
}

test_charset() {
	testname="charset" \
	configure_args="--enable-charset" \
	confbuild
}

test_no_charset() {
	testname="no_charset" \
	configure_args="--disable-charset" \
	confbuild
}

test_X11() {
	testname="X11" \
	configure_args="--with-x" \
	confbuild
}

test_no_X11() {
	testname="no_X11" \
	configure_args="--without-x" \
	confbuild
}

test_vfs() {
	testname="vfs" \
	configure_args="--with-vfs" \
	confbuild
}

test_no_vfs() {
	testname="no_vfs" \
	configure_args="--without-vfs" \
	confbuild
}

test_mcfs() {
	testname="mcfs" \
	configure_args="--with-mcfs" \
	confbuild
}

test_no_mcfs() {
	testname="no_mcfs" \
	configure_args="--without-mcfs" \
	confbuild
}

test_samba() {
	testname="samba" \
	configure_args="--with-samba" \
	confbuild
}

test_no_samba() {
	testname="no_samba" \
	configure_args="--without-samba" \
	confbuild
}

test_ext2undel() {
	testname="ext2undel" \
	configure_args="--with-ext2undel" \
	confbuild
}

test_no_ext2undel() {
	testname="no_ext2undel" \
	configure_args="--without-ext2undel" \
	confbuild
}

test_slang() {
	testname="slang" \
	configure_args="--with-screen=slang" \
	confbuild
}

test_mcslang() {
	testname="mcslang" \
	configure_args="--with-screen=mcslang" \
	confbuild
}

test_ncurses() {
	testname="ncurses" \
	configure_args="--with-screen=ncurses" \
	confbuild
}

test_maintainer() {
	testname="maintainer" \
	configure_args="--enable-maintainer-mode" \
	confbuild
}

test_no_maintainer() {
	testname="no_maintainer" \
	configure_args="--disable-maintainer-mode" \
	confbuild
}

test_i18n() {
	testname="i18n" \
	configure_args="--enable-nls" \
	confbuild
}

test_no_i18n() {
	testname="no_i18n" \
	configure_args="--disable-nls" \
	confbuild
}

test_no_features() {
	testname="no_features" \
	configure_args="${configure_args_disable_all}"
	confbuild
}

test_all_features() {
	testname="all_features" \
	configure_args="${configure_args_enable_all}" \
	confbuild
}

test_warnings1() {
	testname="warnings1" \
	configure_args="${configure_args_enable_all}" \
	GCC_CFLAGS="${gcc_flags_warns2}" \
	confbuild
}

test_warnings2() {
	testname="warnings2" \
	configure_args="${configure_args_enable_all}" \
	GCC_CFLAGS="${gcc_flags_warns2}" \
	confbuild
}

test_all_warnings() {
	testname="all_warnings" \
	configure_args="${configure_args_enable_all} --enable-gcc-warnings" \
	confbuild
}

test_debug() {
	testname="debug" \
	configure_args="${configure_args_enable_all}" \
	GCC_CFLAGS="-ggdb ${gcc_flags_warns2} -O0" \
	confbuild
}

test_profile() {
	testname="profile" \
	configure_args="${configure_args_enable_all}" \
	GCC_CFLAGS="-pg -ggdb" \
	confbuild
}

test_no_optimize() {
	testname="no_optimize" \
	configure_args="${configure_args_enable_all}"
	GCC_CFLAGS="-O0" \
	confbuild
}

test_optimize_level3() {
	testname="optimize_level3" \
	configure_args="${configure_args_enable_all}"
	GCC_CFLAGS="-O3" \
	confbuild
}

test_optimize_size() {
	testname="optimize_size" \
	configure_args="${configure_args_enable_all}"
	GCC_CFLAGS="-Os" \
	confbuild
}

have_check_atleast_version() {
	#local dir ckmaj ckmin ckmic sed_cmd

	sed_cmd="sed -n -e '\$ s,.*\([0-9][0-9]*\).*\$,\1,p'"
	ckmaj=`{ echo "#include <check.h>"; echo "CHECK_MAJOR_VERSION"; } | ${CPP} ${cppflags} -E - 2>/dev/null | eval "${sed_cmd}"`
	ckmin=`{ echo "#include <check.h>"; echo "CHECK_MINOR_VERSION"; } | ${CPP} ${cppflags} -E - 2>/dev/null | eval "${sed_cmd}"`
	ckmic=`{ echo "#include <check.h>"; echo "CHECK_MICRO_VERSION"; } | ${CPP} ${cppflags} -E - 2>/dev/null | eval "${sed_cmd}"`

	if test -z ${ckmaj}${ckmin}${ckmic}; then
		return 1
	elif test ${ckmaj} -gt $1; then
		return 0
	elif test ${ckmaj} -eq $1 && test ${ckmin} -gt $2; then
		return 0
	elif test ${ckmaj} -eq $1 && test ${ckmin} -eq $2 && test ${ckmic} -eq $3; then
		return 0
	else
		return 1
	fi
}

test_check() {
	checkdir=""
	for dir in "/usr" "/usr/local" "/opt/pkg" "${HOME}/pkg"; do
		if cppflags="-I${dir}/include ${CPP_FLAGS}" have_check_atleast_version 0 9 1; then
			checkdir="${dir}"
			break
		fi
	done
	if test -n "${checkdir}"; then
		testname="check" \
		configure_args="${configure_args_enable_all} --with-check=${checkdir}" \
		confbuild
	else
		echo "[`date`] did not find check -- skipping test"
	fi
}

LC_ALL="C"
LC_MESSAGES="C"
export LC_ALL LC_MESSAGES

while test $# -ne 0; do
	case "$1" in
		-basedir)		shift; dir_basedir="$1"; shift;;
		-cvs-repository)		shift; cvs_repository="$1"; shift;;
		-cvs-module)		shift; cvs_module="$1"; shift;;
		-srcdir)		shift; dir_srcdir="$1"; shift;;

		-cleanup)		shift; do_cleanup=yes;;

		-cleansrc)		shift; do_cleansrc=yes;;
		-dont-cleansrc)		shift; do_cleansrc=no;;
		-preconfigure)		shift; do_preconfigure=once;;
		-dont-preconfigure)	shift; do_preconfigure=no;;
		-cleanwork)		shift; do_cleanwork=yes;;
		-dont-cleanwork)	shift; do_cleanwork=no;;
		-configure)		shift; do_configure=yes;;
		-dont-configure)	shift; do_configure=no;;
		-clean)			shift; do_make_clean=yes;;
		-dont-clean)		shift; do_make_clean=no;;
		-build)			shift; do_make_all=yes;;
		-dont-build)		shift; do_make_all=no;;
		-check)			shift; do_make_check=yes;;
		-dont-check)		shift; do_make_check=no;;
		-install)		shift; do_make_install=yes;;
		-dont-install)		shift; do_make_install=no;;

		-fast)			shift; do_cleanwork=no
					do_preconfigure=no
					do_configure=no
					do_make_clean=no
					do_make_all=yes;;
		-nice)			shift; nice_value="$1"; shift;;
		-quiet)			shift; quiet="yes";;
		-enable-gcc-warnings)	shift; enable_gcc_warnings=yes;;
		-disable-gcc-warnings)	shift; enable_gcc_warnings=no;;

		--) shift; break;;
		-) shift; break;;
		-*) usage;;
		*) break;;
	esac
done

if test -z "${dir_basedir}"; then
	dir_basedir="${TMPDIR:-"/tmp"}/${cvs_module}"
fi
if test -z "${dir_srcdir}"; then
	dir_srcdir="${dir_basedir}/src"
fi
dir_prefix="${dir_basedir}/installed"
dir_workdir="${dir_basedir}/build"
dir_logdir="${dir_basedir}/logs"

if test ${do_cleanup} = yes; then
	phase_start "cleaning up everything ..."
	{ rm -rf "${dir_basedir}"
	} && phase_ok || phase_failed
elif test $# -eq 0; then
	all_tests=`sed -n -e 's,^\(test_[a-z0-9_]*\)() .*{,\1,p' $0`
	for i in ${all_tests}; do
		eval "${i}"
	done
else
	for i in "$@"; do
		if test "`type -t test_"${i}"`" = "function"; then
			eval "test_${i}"
		else
			echo "[`date`] test \"$i\" not found." 1>&2
			errors=`expr $errors + 1`
		fi
	done
fi

if test ${errors} -ne 0 || test ${warnings} -ne 0; then
	echo "[`date`] finished with ${errors} errors and ${warnings} warnings."
else
	echo "[`date`] finished successfully."
fi
exit ${errors}
